/********************************************************************/
/*
Modified from pendulum2.cpp ODE demo program.
Control a two link pendulum., with a specified initial condition
*/
/********************************************************************/
/* Include ODE and graphics stuff */

#include "ode/ode.h" /* Dynamics stuff */
#include "drawstuff/drawstuff.h" /* Graphics stuff */
#include "ode-mumble.h" /* CGA stuff to make things clearer */

/********************************************************************/
/********************************************************************/
/* Defines (constants) */

#define MAX_N_BODIES 100 // maximum number of bodies 
#define MAX_N_JOINTS 100 // maximum number of joints

// some constants
#define LENGTH (0.5f)	// length of link
#define WIDTH (0.1f)    // width of link
#define MASS (0.5f)     // mass of link

#define G (9.81f)

#define TIMESTEP (0.01f)

/* Servo gains */
#define SERVO_K0 (10.0f)
#define SERVO_B0 (1.0f)
#define SERVO_K1 (5.0f)
#define SERVO_B1 (1.0f)

/* Score parameters */
#define VISCOUS_FRICTION 0.1
#define STATE_PENALTY 0.1

/* Handy constants */
#define UPPER 0 // upper arm
#define LOWER 1 // forearm
#define SHOULDER 0
#define ELBOW 1
#define XX 0 // indices for XYZ vectors.
#define YY 1
#define ZZ 2

/********************************************************************/
/* Globals */

// ODE objects.
static dWorldID world;         // the simulation object
static dBodyID body[MAX_N_BODIES]; // body IDs get stashed here
static dJointID joint[MAX_N_JOINTS]; // joint IDs get stashed here
  
dReal time = 0.0; // keep track of how much time has passed.

dReal score = 0; // keep track of the score so far.
dReal energy = 0; // keep track of energy of pendulum.

// initial state
float initial_angle[2] = { 0.0f, 0.0f };
float initial_angular_velocity[2] = { 0.0f, 0.0f };

// moment of inertia of pendulum around joint.
dReal I_cm[2] = { 0.0f, 0.0f }; 

float g_servo_shoulder[4] = {0.0f}; //{ 131.53f, 40.95f,  .68f, .21f };
float g_servo_elbow[4] = {0.0f}; //{ 40.92f, 17.29f, .21f, .08f };

float g_balance_state[4] = { M_PI, 0.0f, 0.0f, 0.0f };

/********************************************************************/

dReal square(dReal x)
{
  return x*x;
}

/********************************************************************/
// start(): set things up

static void start()
{
  float ic[2], is[2];
  dMatrix3 R;
  dReal elbow[3], elbowd[3];
  dReal cm[2][3];
  dReal cmd[2][3];
  dReal lower_ang;
  dReal lower_ang_vel;

  // create world
  world = dWorldCreate();
  dWorldSetGravity ( world, 0, 0, -G ); // gravity is down along Z axis

  /* set up mass, cm location, and moment of inertia for links */
  dMass m;
  dMassSetBox (&m,1,WIDTH,WIDTH,LENGTH);
  dMassAdjust (&m,MASS);
  // Print out inertial parameters of a link.
  printf( "Inertial parameters: %g; %g %g %g %g;\n", m.mass,
	  m.c[0], m.c[1], m.c[2], m.c[3] );
  printf( "%g %g %g %g\n", m.I[0], m.I[1], m.I[2], m.I[3] );
  printf( "%g %g %g %g\n", m.I[4], m.I[5], m.I[6], m.I[7] );
  printf( "%g %g %g %g\n", m.I[8], m.I[9], m.I[10], m.I[11] );
  I_cm[UPPER] = m.I[5];
  I_cm[LOWER] = m.I[5];

  // set up bodies
  body[UPPER] = dBodyCreate (world);
  body[LOWER] = dBodyCreate (world);
  dBodySetMass (body[UPPER],&m);
  dBodySetMass (body[LOWER],&m);
  dBodySetPosition( body[UPPER], 0.0, 0.0, 3*LENGTH/2 );
  dBodySetPosition( body[LOWER], 0.0, 0.0, LENGTH/2 );

  // create joints: shoulder
  joint[SHOULDER] = dJointCreateHinge( world, 0 );
  dJointAttach( joint[SHOULDER], body[UPPER], 0 );
  dJointSetHingeAnchor( joint[SHOULDER], 0.0, 0.0, 2*LENGTH );
  dJointSetHingeAxis( joint[SHOULDER], 0.0, 1.0, 0.0 );

  // create joints: elbow
  joint[ELBOW] = dJointCreateHinge( world, 0 );
  dJointAttach( joint[ELBOW], body[LOWER], body[UPPER] );
  dJointSetHingeAnchor( joint[ELBOW], 0.0, 0.0, LENGTH );
  dJointSetHingeAxis( joint[ELBOW], 0.0, 1.0, 0.0 );

  // set up view point
  static float xyz[3] = {0.0f,2.0f,1.0f};
  static float hpr[3] = {-90.0000f,0.0000f,0.0000f};
  dsSetViewpoint (xyz,hpr);

  /* Now set up initial position and velocity */
  ic[SHOULDER] = cos((double) initial_angle[SHOULDER] );
  is[SHOULDER] = sin((double) initial_angle[SHOULDER] );
  lower_ang = initial_angle[SHOULDER] + initial_angle[ELBOW];
  ic[ELBOW] = cos((double) lower_ang );
  is[ELBOW] = sin((double) lower_ang );
  lower_ang_vel = initial_angular_velocity[SHOULDER]
    + initial_angular_velocity[ELBOW];
  cm[UPPER][XX] = -is[SHOULDER]*LENGTH/2;
  cm[UPPER][YY] = 0.0;
  cm[UPPER][ZZ] = LENGTH*(2 - ic[SHOULDER]/2);
  dBodySetPosition( body[UPPER], cm[UPPER][XX], cm[UPPER][YY], cm[UPPER][ZZ] );
  elbow[XX] = -is[SHOULDER]*LENGTH;
  elbow[YY] = 0;
  elbow[ZZ] = LENGTH*(2 - ic[SHOULDER]);
  dRFromAxisAndAngle( R, 0.0, 1.0, 0.0, initial_angle[SHOULDER] );
  dBodySetRotation( body[UPPER], R );
  cmd[UPPER][XX] = -initial_angular_velocity[SHOULDER]*LENGTH*ic[SHOULDER]/2;
  cmd[UPPER][YY] = 0.0;
  cmd[UPPER][ZZ] = initial_angular_velocity[SHOULDER]*LENGTH*is[SHOULDER]/2;
  dBodySetLinearVel( body[UPPER], 
		     cmd[UPPER][XX], cmd[UPPER][YY], cmd[UPPER][ZZ] );
  elbowd[XX] = -initial_angular_velocity[SHOULDER]*LENGTH*ic[SHOULDER];
  elbowd[YY] = 0;
  elbowd[ZZ] = initial_angular_velocity[SHOULDER]*LENGTH*is[SHOULDER];
  dBodySetAngularVel( body[UPPER], 0.0, initial_angular_velocity[SHOULDER], 0.0 );
  
  cm[LOWER][XX] = -is[ELBOW]*LENGTH/2 + elbow[XX];
  cm[LOWER][YY] = elbow[YY];
  cm[LOWER][ZZ] = LENGTH*(- ic[ELBOW]/2) + elbow[ZZ];
  dBodySetPosition( body[LOWER], cm[LOWER][XX], cm[LOWER][YY], cm[LOWER][ZZ] );
  dRFromAxisAndAngle( R, 0.0, 1.0, 0.0, 
		      initial_angle[SHOULDER] + initial_angle[ELBOW] );
  dBodySetRotation( body[LOWER], R );
  cmd[LOWER][XX] = -lower_ang_vel*LENGTH*ic[ELBOW]/2
    + elbowd[XX];
  cmd[LOWER][YY] = elbowd[YY];
  cmd[LOWER][ZZ] = lower_ang_vel*LENGTH*is[ELBOW]/2
    + elbowd[ZZ];
  dBodySetLinearVel( body[LOWER], 
		     cmd[LOWER][XX], cmd[LOWER][YY], cmd[LOWER][ZZ] );
  dBodySetAngularVel( body[LOWER], 0.0, lower_ang_vel, 0.0 );

  energy = 0.5*I_cm[UPPER]*square(initial_angular_velocity[SHOULDER])
    + 0.5*MASS*(square(cmd[UPPER][XX]) + square(cmd[UPPER][YY])
		+ square(cmd[UPPER][ZZ]))
    + MASS*G*(cm[UPPER][ZZ] - 3*LENGTH/2)
    + 0.5*I_cm[LOWER]*square(lower_ang_vel)
    + 0.5*MASS*(square(cmd[LOWER][XX]) + square(cmd[LOWER][YY])
		+ square(cmd[LOWER][ZZ]))
    + MASS*G*(cm[LOWER][ZZ] - LENGTH/2);

  printf( "Initial state: %g %g %g %g; energy: %g\n", 
	  initial_angle[SHOULDER], initial_angle[ELBOW],
	  initial_angular_velocity[SHOULDER], initial_angular_velocity[ELBOW],
	  energy );

  printf ("Simulating ...\n");
}

/********************************************************************/
// called when a key pressed

static void command (int cmd)
{
  // don't handle user input yet.
}

/********************************************************************/
/* read and joint, and handle +/-PI offset */

dReal
read_angle_and_fix_offset( dJointID joint, dReal *last_angle,
			   dReal *joint_offset )
{
  dReal raw_angle, angle;

  raw_angle = dJointGetHingeAngle( joint );
  angle = raw_angle + *joint_offset;

  if ( *last_angle - angle > M_PI )
    *joint_offset += 2*M_PI;
  else if ( angle - *last_angle > M_PI )
    *joint_offset -= 2*M_PI;
  
  angle = raw_angle + *joint_offset;
  *last_angle = angle;
  return angle;
}

/********************************************************************/

// simulation loop

static void simLoop (int pause)
{
  static int count = 0;
  dReal angle_desired[2] = { M_PI, 0.0 };
  dReal angular_velocity_desired[2] = { 0.0, 0.0 };
  dReal angle_offset[2] = { 0.0, 0.0 };
  static dReal last_angle[2] = { 0.0, 0.0 };
  //dReal k[2] = { SERVO_K0, SERVO_K1 };
  //dReal b[2] = { SERVO_B0, SERVO_B1 };
  dReal angle[2];
  dReal angular_velocity[2];
  dReal torque[2];
  dReal *cm[2];
  dReal *cmd[2];

  if (!pause)
    {
      
      /* Get status of joints */
      angle[SHOULDER] = read_angle_and_fix_offset( joint[SHOULDER], 
						   &last_angle[SHOULDER],
						   &angle_offset[SHOULDER] );
      angle[ELBOW] = read_angle_and_fix_offset( joint[ELBOW], 
						&last_angle[ELBOW],
						&angle_offset[ELBOW] );
      cm[UPPER] = (double *) dBodyGetPosition( body[UPPER] );
      cmd[UPPER] = (double *) dBodyGetLinearVel( body[UPPER] );
      cm[LOWER] = (double *) dBodyGetPosition( body[LOWER] );
      cmd[LOWER] = (double *) dBodyGetLinearVel( body[LOWER] );

      angular_velocity[SHOULDER] = dJointGetHingeAngleRate( joint[SHOULDER] );
      angular_velocity[ELBOW] = dJointGetHingeAngleRate( joint[ELBOW] );
      //torque[SHOULDER] = -k[SHOULDER]*( angle[SHOULDER] - angle_desired[SHOULDER] ) - b[SHOULDER]*angular_velocity[SHOULDER];
      //torque[ELBOW] = -k[ELBOW]*( angle[ELBOW] - angle_desired[ELBOW] ) - b[ELBOW]*angular_velocity[ELBOW];

      torque[SHOULDER] = -g_servo_shoulder[0]*( angle[SHOULDER] - angle_desired[SHOULDER] ) 
                         -g_servo_shoulder[1]*( angle[ELBOW] - angle_desired[ELBOW] )
                         -g_servo_shoulder[2]*( angular_velocity[SHOULDER] - angular_velocity_desired[SHOULDER] )
                         -g_servo_shoulder[3]*( angular_velocity[ELBOW] - angular_velocity_desired[ELBOW] );

      torque[ELBOW] =    -g_servo_elbow[0]*( angle[SHOULDER] - angle_desired[SHOULDER] ) 
                        -g_servo_elbow[1]*( angle[ELBOW] - angle_desired[ELBOW] )
                         -g_servo_elbow[2]*( angular_velocity[SHOULDER] - angular_velocity_desired[SHOULDER] )
                         -g_servo_elbow[3]*( angular_velocity[ELBOW] - angular_velocity_desired[ELBOW] );

      dJointAddHingeTorque( joint[SHOULDER],
			    torque[SHOULDER] 
			    - VISCOUS_FRICTION*angular_velocity[SHOULDER] ); 
      dJointAddHingeTorque( joint[ELBOW],
			    torque[ELBOW] 
			    - VISCOUS_FRICTION*angular_velocity[ELBOW] ); 

      dWorldStep( world, TIMESTEP );
      time += TIMESTEP;
      score += square(torque[SHOULDER])*TIMESTEP 
	+ square(torque[ELBOW])*TIMESTEP 
	+ STATE_PENALTY*TIMESTEP
	*square(angle[SHOULDER] - angle_desired[SHOULDER])
	+ STATE_PENALTY*TIMESTEP
	*square(angle[ELBOW] - angle_desired[ELBOW] );

      energy = 0.5*I_cm[UPPER]*square(angular_velocity[SHOULDER])
	+ 0.5*MASS*(square(cmd[UPPER][XX]) + square(cmd[UPPER][YY])
		    + square(cmd[UPPER][ZZ]))
	+ MASS*G*(cm[UPPER][ZZ] - 3*LENGTH/2)
	+ 0.5*I_cm[LOWER]*square(angular_velocity[SHOULDER]
				 + angular_velocity[ELBOW])
	+ 0.5*MASS*(square(cmd[LOWER][XX]) + square(cmd[LOWER][YY])
		    + square(cmd[LOWER][ZZ]))
	+ MASS*G*(cm[LOWER][ZZ] - LENGTH/2);

      count++;
      if ( (count % 100) == 0 )
	printf( "%g %g %g %g %g %g %g %g\n", time,
		angle[SHOULDER], angle[ELBOW],
		angular_velocity[SHOULDER], angular_velocity[ELBOW], 
		torque[SHOULDER], torque[ELBOW],
		score );
    }

  /* Graphics */
  dReal sides1[3] = {WIDTH,WIDTH,LENGTH};
  dsSetTexture (DS_WOOD);
  dsSetColor (1,1,0);
  dsDrawBox (dBodyGetPosition(body[0]),dBodyGetRotation(body[0]),sides1);
  dsDrawBox (dBodyGetPosition(body[1]),dBodyGetRotation(body[1]),sides1);
}

/********************************************************************/

int main (int argc, char **argv)
{

  CALLBACK_STUFF;
  FILE *fileK;
  initial_angle[SHOULDER] = 0.0;
  initial_angle[ELBOW] = 0.0;
  initial_angular_velocity[SHOULDER] = 0.0; 
  initial_angular_velocity[ELBOW] = 0.0; 
  char   constants[256];

  /* This doesn't seem to work under windows */
  if ( argc > 1 )
    sscanf( argv[1], "%g", &(initial_angle[SHOULDER]) );
  if ( argc > 2 )
    sscanf( argv[2], "%g", &(initial_angle[ELBOW]) );
  if ( argc > 3 )
    sscanf( argv[3], "%g", &(initial_angular_velocity[SHOULDER]) );
  if ( argc > 4 )
    sscanf( argv[4], "%g", &(initial_angular_velocity[ELBOW]) );


  //read in constants K
  fileK = fopen("K", "r");
  if (fileK == NULL)
  {
    printf("Constant file \"K\" is missing\n");
    exit(1);
  }

  //read the first row of the shoulder constants
  fgets(constants, 256, fileK);
  sscanf(constants, "%g %g %g %g", &(g_servo_shoulder[0]), &(g_servo_shoulder[1]), &(g_servo_shoulder[2]), &(g_servo_shoulder[3]));

  //read arm constants
  fgets(constants, 256, fileK);
  sscanf(constants, "%g %g %g %g", &(g_servo_elbow[0]), &(g_servo_elbow[1]), &(g_servo_elbow[2]), &(g_servo_elbow[3]));

  printf("Servo Constants K\n");
  printf("Shoulder:  %6g %6g %6g %6g\n", g_servo_shoulder[0], g_servo_shoulder[1], g_servo_shoulder[2], g_servo_shoulder[3]);
  printf("Elbow:  %6g %6g %6g %6g\n", g_servo_elbow[0], g_servo_elbow[1], g_servo_elbow[2], g_servo_elbow[3]);

  fclose(fileK);

  // run simulation
  dsSimulationLoop( argc, argv, /* command line arguments */
		    352, 288, /* window size? */
		    &fn ); /* graphics info */

  // clean up stuff (not critical, we are killing process anyway 
  dWorldDestroy( world );

  return 0;
}

/********************************************************************/
