/************************************************************************/
/*
dynamics for two link pendulum
*/
/************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "dynamics-2-link.h"

#define LENGTH (0.5f)	// length of link
#define WIDTH (0.1f)    // width of link
#define MASS (0.5f)     // mass of link
#define GRAVITY 9.81
#define VISCOUS_FRICTION 0.1 // viscous friction at each joint

#define TIMESTEP 0.01 // Timestep of integrator

/* Score parameters */
#define STATE_PENALTY 0.1

/**********************************************************************/
/* Globals */

float the_state[N_STATE_DIMENSIONS] = { 0.0, 0.0, 0.0, 0.0 };
float the_time = 0;
float the_desired_state[N_STATE_DIMENSIONS] = { M_PI, 0, 0, 0 };

/************************************************************************/

static float mysquare(float i)
{
  return i*i;
}

/**********************************************************************/

int init_dynamics()
{
  return 1;
}

/**********************************************************************/

int set_state( float time, float *state )
{
  int i; 

  the_time = time;
  for ( i = 0; i < N_STATE_DIMENSIONS; i++ )
    the_state[i] = state[i];
  return 1;
}

/**********************************************************************/

/*
#define m0 MASS
#define m1 MASS
#define l0cm (LENGTH/2)
#define l1cm (LENGTH/2)
#define l0 LENGTH
#define G GRAVITY
#define I0yy (MASS*(LENGTH*LENGTH + WIDTH*WIDTH)/12)
#define I1yy (MASS*(LENGTH*LENGTH + WIDTH*WIDTH)/12)
*/

/**********************************************************************/

int integrate_one_step( float *action, float *next_state )
{
  int i;
  float a0 = 0;
  float a1 = 0;
  float a0d = 0;
  float a1d = 0;
  float torque0 = 0;
  float torque1 = 0;
  /* Slightly faster to have these as variables than defines. Go figure */
  float m0 = MASS;
  float m1 = MASS;
  float l0cm = (LENGTH/2);
  float l1cm = (LENGTH/2);
  float l0 = LENGTH;
  float G = GRAVITY;
  float I0yy = (MASS*(LENGTH*LENGTH + WIDTH*WIDTH)/12);
  float I1yy = (MASS*(LENGTH*LENGTH + WIDTH*WIDTH)/12);
  float a0dd = 0;
  float a1dd = 0;
  float a, b, c, d, e, f;
  float determinant;
  float new_a0d, new_a1d;
  float sa0, ca0, sa1, ca1;

  a0 = the_state[0];
  a1 = the_state[1];
  a0d = the_state[2];
  a1d = the_state[3];

  torque0 = action[0];
  torque1 = action[1];

  sa0 = sinf( a0 );
  ca0 = cosf( a0 );
  sa1 = sinf( a1 );
  ca1 = cosf( a1 );

  a = I0yy + I1yy + l0cm*l0cm*m0 + l0*l0*m1 + l1cm*l1cm*m1
    + 2*l0*l1cm*m1*ca1;
  b = I1yy + l1cm*l1cm*m1 + l0*l1cm*m1*ca1;
  e = torque0 - VISCOUS_FRICTION*a0d
    - ( G*l0cm*m0*sa0 + G*l0*m1*sa0 + G*l1cm*m1*ca1*sa0
	- 2*a0d*a1d*l0*l1cm*m1*sa1 - a1d*a1d*l0*l1cm*m1*sa1
	+ G*l1cm*m1*ca0*sa1 );
  
  c = I1yy + l1cm*l1cm*m1 + l0*l1cm*m1*ca1;
  d = I1yy + l1cm*l1cm*m1;
  f = torque1 - VISCOUS_FRICTION*a1d
    - ( G*l1cm*m1*ca1*sa0 
	+ a0d*a0d*l0*l1cm*m1*sa1 + G*l1cm*m1*ca0*sa1 );

  determinant = a*d - b*c;
  a0dd = (d*e - b*f)/determinant;
  a1dd = (-c*e + a*f)/determinant;

  new_a0d = a0d + a0dd*TIMESTEP;
  a0 += (new_a0d + a0d)*TIMESTEP/2;
  a0d = new_a0d;
  new_a1d = a1d + a1dd*TIMESTEP;
  a1 += (new_a1d + a1d)*TIMESTEP/2;
  a1d = new_a1d;
  the_time += TIMESTEP;
  the_state[0] = a0;
  the_state[1] = a1;
  the_state[2] = a0d;
  the_state[3] = a1d;
  if ( next_state != NULL )
    {
      for( i = 0; i < N_STATE_DIMENSIONS; i++ )
	next_state[i] = the_state[i];
    }
}

/**********************************************************************/

float one_step_cost( float *state, float *action )
{
  float s0, s1;

  s0 = state[0] - the_desired_state[0];
  s1 = state[1] - the_desired_state[1];
  return action[0]*action[0]*TIMESTEP 
    + action[1]*action[1]*TIMESTEP 
    + STATE_PENALTY*TIMESTEP*s0*s0
    + STATE_PENALTY*TIMESTEP*s1*s1;
}

/**********************************************************************/

